/* File CGPROCS.C Conjugate gradient procedures */

   /* Calling programs must supply the objective function (the_obj_fn)  */
   /*   and the gradient of the objective function.  Also, calling programs */
   /*   supply an allocated and initialized d_vector.  */

#ifndef CGPROCS_C
#define CGPROCS_C

#define debug

#include <stdlib.h>
#include <conio.h>
#include <string.h>
#include <math.h>

#include "ut.h"
#include "cg.h"

short conjugate_gradient (float (*obj_fn)(d_vector w),
     void (*gradient_of_obj_fn)(d_vector w,d_vector grad),
     d_vector weights,int no_of_wts,float error_tol,float *opt_obj_value);

/*  Implementation  */

const float cg_zero_check = 1e-8;

double dot_prod (d_vector v1,d_vector v2,int start,int len) {
   int i;
   double sum = 0.0;
   for (i = 0;i < len;i++)
      sum += v1[i+start]*v2[i+start];
   return sum;
   }  /*  function  */

void minus (d_vector v,d_vector neg_v,int start,int len) {
     int i;
     for (i = 0;i<len;i++)
	  neg_v[i + start] = -v[i + start];
     }  /*  function  */

void display_vect (char *msg,d_vector v,int start,int len) {
     int i,disp_n,room;
     room = 80 - strlen (msg);
     disp_n = room/8;
     if (len < disp_n) disp_n = len;
     cputs (msg);
     for (i = 0;i<disp_n;i++)
	 cprintf ("%7.5lf ",v[i+start]);
     return;
     }  /*  proc  */

void form_new_vector (float a,d_vector w1,d_vector w2,d_vector v,
     int start,int len) {
     int i;
     for (i = 0;i<len;i++)
	  v[i+start] = w1[i+start] + a * w2[i+start];
     return;
     }  /* proc  */

double vect_norm (d_vector v,int start,int len) {
   int i;
   double sum = 0.0;

   for (i = 0;i<len;i++)
       sum += v[i+start]*v[i+start];
   return sqrt(sum);
   }  /*  function  */

void unit_vector (d_vector v,d_vector u,int start,int len) {
     double denom;
     int i;
     denom = vect_norm (v,start,len);
     if (denom < cg_zero_check) denom = cg_zero_check;
     for (i = 0;i<len;i++)
	 u[i+start] = v[i+start]/denom;
     return;
     }

double find_beta (d_vector grad,d_vector new_grad,int start,int len) {
   double denom;
   denom = dot_prod(grad,grad,start,len);
   if (denom < cg_zero_check) denom = cg_zero_check;
   return dot_prod (new_grad,new_grad,start,len)/denom;
   } /* function */

void find_min (d_vector weights,d_vector direction_vector,int start,int len,
	       float (*obj_fn)(d_vector w),
	       float start_range,double *min_x,float *min_f) {
/* Brute force! */
    const float f_zero_toler = 1e-8, x_zero_toler = 1e-4;
    int i,no_of_pts = 100;
    float opt_value,value,opt_x,x,xinc,xmin,range;
    d_vector v;
    v = allocate_d_vector (1,len);
    range = start_range;
    do {
       xmin = -range;
       xinc = 2*range/no_of_pts;
       x = xmin;
       form_new_vector (x,weights,direction_vector,v,start,len);
       opt_value = obj_fn (v);
       opt_x = x;
       for (i = 1;i<=no_of_pts;i++) {
	   x += xinc;
	   form_new_vector (x,weights,direction_vector,v,start,len);
	   value = obj_fn (v);
	   if (value < opt_value) {
	      opt_x = x;
	      opt_value = value;
	      }
	   }  /* i */
	*min_x = opt_x;
	*min_f = opt_value;
	range *= 0.1;
    /* Repeat until either a sizable move (x) is found, or the func is zero */
	} while ((fabs (*min_x) < x_zero_toler)&&(opt_value > f_zero_toler)&&
		(range > x_zero_toler));
    if (range < x_zero_toler)
       cprintf ("\r\n**** find_min can't find a min in this direction.");
    free_d_vector (v,1);
    }  /* proc */

short conjugate_gradient (float (*obj_fn)(d_vector w),
     void (*gradient_of_obj_fn)(d_vector w,d_vector grad),
     d_vector weights,int no_of_wts,float error_tol,float *opt_obj_value) {

     d_vector gradient,new_gradient,
	      neg_new_gradient,new_vector,
	      neg_gradient,new_weights,
	      direction_vector;
     long int iter = 0;
     double alpha,beta,alpha_start,alpha_end,the_error;
     float obj_value,step_size;
     step_size = vect_norm (weights,1,no_of_wts);  /* starting value */
     gradient = allocate_d_vector (1,no_of_wts);
     neg_gradient = allocate_d_vector (1,no_of_wts);
     new_weights = allocate_d_vector (1,no_of_wts);
     direction_vector = allocate_d_vector (1,no_of_wts);
     new_gradient = allocate_d_vector (1,no_of_wts);
     neg_new_gradient = allocate_d_vector (1,no_of_wts);
     new_vector = allocate_d_vector (1,no_of_wts);
     clrscr();
     display_vect ("\r\nInit Wts: ",weights,1,no_of_wts);
     gradient_of_obj_fn (weights,gradient);
     display_vect ("\r\nGradient: ",gradient,1,no_of_wts);
     minus (gradient,neg_gradient,1,no_of_wts);
     unit_vector (neg_gradient,direction_vector,1,no_of_wts);
     the_error = vect_norm (gradient,1,no_of_wts);
     display_vect ("\r\nDir Vect: ",direction_vector,1,no_of_wts);
     find_min (weights,direction_vector,1,no_of_wts,obj_fn,step_size,
	  &alpha,&obj_value);
     cprintf ("\r\nalpha = %7.5lf, obj value = %8.5f, grad norm = %10.8lf",
	     alpha,obj_value,the_error);
     cprintf ("\r\nError_tol: %10.8f",error_tol);
     form_new_vector (alpha,weights,direction_vector,new_weights,1,no_of_wts);
     copy_d_vector (new_weights,weights,1,no_of_wts);
     display_vect("\r\nWeights: ",weights,1,no_of_wts);
     cputs ("\r\nPress key: ");
     getch();
     while ((obj_value > error_tol) && !kbhit()) {
	iter++;
	gradient_of_obj_fn (weights,new_gradient);
	display_vect ("\r\nNew Grad: ",new_gradient,1,no_of_wts);
	minus (new_gradient,neg_new_gradient,1,no_of_wts);
	if ((iter % no_of_wts)==0)
	   unit_vector (neg_new_gradient,direction_vector,1,no_of_wts);
	else {
	   beta = find_beta (gradient,new_gradient,1,no_of_wts);
	   cprintf ("\r\nBeta = %7.5lf",beta);
	   form_new_vector (beta,neg_new_gradient,direction_vector,
		    new_vector,1,no_of_wts);
	   display_vect("\r\nNew Vect: ",new_vector,1,no_of_wts);
	   unit_vector (new_vector,direction_vector,1,no_of_wts);
	   display_vect("\r\nNew Direct: ",direction_vector,1,no_of_wts);
	   }  /*  else  */
	copy_d_vector (new_gradient,gradient,1,no_of_wts);
	the_error = vect_norm (gradient,1,no_of_wts);
	if (the_error > 1e-6)
	   step_size = 5*(obj_value)/the_error;  /* '5*y divided by the slope' */
	else step_size = 0.5 * step_size;
	find_min (weights,direction_vector,1,no_of_wts,obj_fn,step_size,
	   &alpha,&obj_value);
	cprintf ("\r\nStep size: %8.6f",step_size);
	cprintf ("\r\nalpha = %8.5lf, obj_value = %8.5lf",alpha,obj_value);
	form_new_vector (alpha,weights,direction_vector,new_weights,
	   1,no_of_wts);
	copy_d_vector (new_weights,weights,1,no_of_wts);
	display_vect ("\r\nNew Wts: ",weights,1,no_of_wts);
	cprintf ("\r\nIter: %5ld, Grad Norm: %8.6lf,  Obj Value: %8.6lf"
	    "\r\n\n\n",iter,the_error,obj_value);
	} /* while */
     *opt_obj_value = obj_value;
     if (kbhit()) getch();
     free_d_vector (gradient,1);
     free_d_vector (neg_gradient,1);
     free_d_vector (new_weights,1);
     free_d_vector (direction_vector,1);
     free_d_vector (new_gradient,1);
     free_d_vector (neg_new_gradient,1);
     free_d_vector (new_vector,1);
     return 1;
     }  /* proc */

#endif
